package DEFLATE.SimpleDeflate;

import Bytes.Bytes.ByteArray;

import static Bytes.Bytes.Bytes.*;
import static java.lang.Math.*;

public class SimpleDeflate {
    public static ByteArray Pack(ByteArray data, double level){
        ByteArray output, trimmed;
        double read, i, j, write, start, len, beststart, bestlen, maxforward, size, plain, maxlength;
        boolean done;

        // 5-65536
        size = level - 1d;
        maxlength = 255d;

        output = CreateByteArray(ceil(ByteArrayLength(data)*1.1));

        write = 0d;
        read = 0d;
        plain = 0d;
        for(; read < ByteArrayLength(data); ){
            // Search
            bestlen = 0d;
            beststart = 0d;
            // Lower limit at most 0.
            start = max(0d, read - size);
            for (i = read - 1d; i >= start; i = i - 1d) {
                // Maximum forward is limited by 1) size, 2) data length or 3) distance to down to search start (i)
                maxforward = min(maxlength, min(ByteArrayLength(data) - read, read - i));
                len = 0d;
                done = false;
                for (j = 0; j < maxforward && !done; j = j + 1d) {
                    if (GetByte(data, i + j) == GetByte(data, read + j)) {
                        len = len + 1d;
                    } else {
                        done = true;
                    }
                }

                if (len > bestlen) {
                    bestlen = len;
                    beststart = read - i;
                }
            }

            // Write data.
            if(read + 1d == ByteArrayLength(data)){
                plain = plain + 1d;
                read = read + 1d;
            }

            if(bestlen >= 4d || plain == 254d || read == ByteArrayLength(data)) {
                if (plain > 0d) {
                    SetByte(output, write, plain);
                    write = write + 1d;
                    for (i = 0d; i < plain; i = i + 1d) {
                        SetByte(output, write, GetByte(data, read - plain + i));
                        write = write + 1d;
                    }
                }
                plain = 0d;
            }

            if(bestlen >= 4d) {
                if(beststart > 255d) {
                    SetByte(output, write + 0d, 0d);
                    SetByte(output, write + 1d, floor(beststart / 256d));
                    SetByte(output, write + 2d, beststart % 256d);
                    SetByte(output, write + 3d, bestlen);
                    write = write + 4d;
                }else{
                    SetByte(output, write + 0d, 255d);
                    SetByte(output, write + 1d, beststart);
                    SetByte(output, write + 2d, bestlen);
                    write = write + 3d;
                }

                read = read + bestlen;
            }else{
                plain = plain + 1d;
                read = read + 1d;
            }
        }

        trimmed = CreateByteArray(write);

        for(i = 0d; i < write; i = i + 1d){
            SetByte(trimmed, i, GetByte(output, i));
        }

        return trimmed;
    }

    public static ByteArray Unpack(ByteArray data){
        ByteArray output;
        double read, write, i, dist, length, pos, start, unpackedLength;

        // Compute length
        unpackedLength = 0d;
        read = 0d;
        for(; read < ByteArrayLength(data); ){
            length = GetByte(data, read);

            if(length == 0d || length == 255d){
                if(length == 0d) {
                    unpackedLength = unpackedLength + GetByte(data, read + 3d);
                    read = read + 4d;
                }else{
                    unpackedLength = unpackedLength + GetByte(data, read + 2d);
                    read = read + 3d;
                }
            }else{
                read = read + 1d;
                unpackedLength = unpackedLength + length;
                read = read + length;
            }
        }

        // Allocate result memory.
        output = CreateByteArray(unpackedLength);

        // Unpack
        write = 0d;
        read = 0d;
        for(; read < ByteArrayLength(data); ){
            length = GetByte(data, read);

            if(length == 0d || length == 255d){
                if(length == 0d) {
                    dist = GetByte(data, read + 1d) * 256d;
                    dist = dist + GetByte(data, read + 2d);
                    length = GetByte(data, read + 3d);
                    read = read + 4d;
                }else{
                    dist = GetByte(data, read + 1d);
                    length = GetByte(data, read + 2d);
                    read = read + 3d;
                }

                start = write - dist;

                for(i = 0; i < length; i = i + 1d){
                    pos = start + i;
                    SetByte(output, write, GetByte(output, pos));
                    write = write + 1d;
                }
            }else{
                read = read + 1d;
                for(i = 0d; i < length; i = i + 1d){
                    SetByte(output, write, GetByte(data, read));
                    write = write + 1d;
                    read = read + 1d;
                }
            }
        }

        return output;
    }

    public static void delete(Object object) {
        // Java has garbage collection.
    }
}
